#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

import argparse
import logging
import os
import pathlib
import tarfile
from contextlib import contextmanager
from pathlib import Path
from shutil import unpack_archive
from tempfile import TemporaryDirectory

description = """Flot is a tool to easily build multiple packages (wheel
and sdist) from a single repo without having to create a subdir or another repo
for each package.
"""

__version__ = "0.6.0"


log = logging.getLogger(__name__)


def main(argv=None):
    parser = argparse.ArgumentParser()
    parser.description = description
    here = Path.cwd()
    parser.add_argument(
        "--pyproject",
        type=Path,
        nargs="?",
        default=here / "pyproject.toml",
        help="pyproject.toml file path. Default: <current directory>/pyproject.toml",
    )

    parser.add_argument(
        "--output-dir",
        dest="output_dir",
        default=here / "dist",
        type=Path,
        nargs="?",
        help="Output directory. Default: <current directory>/dist/",
    )

    parser.add_argument(
        "--wheel",
        action="store_true",
        help="Build a wheel. Default if no option selected.",
    )

    parser.add_argument(
        "--sdist",
        action="store_true",
        help="Build an sdist",
    )

    parser.add_argument(
        "--wheel-tag",
        dest="wheel_tag",
        default="py3-none-any",
        type=str,
        nargs="?",
        help="Optional wheel tag. Default: py3-none-any",
    )

    parser.add_argument(
        "-v", "--version", action="version", version=f"flot {__version__}"
    )

    parser.add_argument("--debug", action="store_true", help=argparse.SUPPRESS)

    args = parser.parse_args(argv)

    enable_logging(level=logging.DEBUG if args.debug else logging.INFO)

    log.debug("Parsed arguments %r", args)

    build_wheel = args.wheel
    build_sdist = args.sdist

    print("Building from pyproject:", args.pyproject)
    print("output in :", args.output_dir)

    from . import sdist
    from . import wheel

    if not (build_wheel or build_sdist):
        build_wheel = True

    if build_sdist:
        sdn = sdist.make_sdist(
            pyproject_file=args.pyproject,
            output_dir=args.output_dir,
        )
        if build_wheel:
            with unpacked_tarball(sdn) as tmpdir:
                sdist_dir = Path(tmpdir)
                wheel.make_wheel(
                    pyproject_file=sdist_dir / "pyproject.toml",
                    output_dir=args.output_dir,
                    wheel_tag=args.wheel_tag,
                )
                print("using wheel tag:", args.wheel_tag)
                print("from unpacked sdist:", tmpdir)

    else:
        wheel.make_wheel(
            pyproject_file=args.pyproject,
            output_dir=args.output_dir,
            wheel_tag=args.wheel_tag,
        )
        print("using wheel tag:", args.wheel_tag)


def enable_logging(level=logging.INFO):
    handler = logging.StreamHandler()
    logging.root.addHandler(handler)
    logging.root.setLevel(level)


@contextmanager
def unpacked_tarball(path):
    tf = tarfile.open(str(path))
    with TemporaryDirectory() as tmpdir:
        tf.extractall(tmpdir)
        files = os.listdir(tmpdir)
        assert len(files) == 1, files
        yield os.path.join(tmpdir, files[0])



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

from pathlib import Path
from shutil import copytree

import pytest

data_dir = Path(__file__).parent / "data"


@pytest.fixture
def copy_test_data(tmp_path):
    """Copy a subdirectory from the data dir to a temp dir"""

    def copy(dirname):
        dst = tmp_path / dirname
        copytree(str(data_dir / dirname), str(dst))
        return dst

    return copy


@pytest.fixture
def get_test_data():
    """Return a Path in the data dir"""

    def _get_path(path):
        return str(data_dir / path)

    return _get_path



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

import logging
from pathlib import Path

import pytest

from flot import config

data_dir = Path(__file__).parent / "data"


def test_load_basic():
    project_info = config.read_pyproject_file(data_dir / "basic" / "my-pyproject.toml")
    expected = {
        "entrypoints": {},
        "excludes": ["**/.git/*", "**/.hg/*"],
        "includes": ["**/*"],
        "metadata": {
            "name": "package1",
            "requires_dist": [],
            "version": "1.0.0",
            "summary": "foo",
        },
        "metadata_files": [],
        "referenced_files": [],
        "reqs_by_extra": {},
        "sdist_extra_excludes": ["**/.git/*", "**/.hg/*"],
        "sdist_extra_includes": [],
        "editable_paths": [],
        "wheel_path_prefixes_to_strip": [],
    }
    assert project_info.to_dict() == expected


def test_toml_with_entry_points():
    project_info = config.read_pyproject_file(
        data_dir / "entrypoints_valid" / "pyproject.toml"
    )
    expected = {
        "editable_paths": [],
        "entrypoints": {
            "console_scripts": {"pkg_script": "package1:main"},
            "myplugins": {"package1": "package1:main"},
        },
        "excludes": ["**/.git/*", "**/.hg/*"],
        "includes": ["module1.py"],
        "metadata": {
            "name": "package1",
            "requires_dist": [],
            "summary": "module1 desc",
            "version": "1.0",
        },
        "metadata_files": [],
        "referenced_files": [],
        "reqs_by_extra": {},
        "sdist_extra_excludes": ["**/.git/*", "**/.hg/*"],
        "sdist_extra_includes": [],
        "wheel_path_prefixes_to_strip": [],
    }
    assert project_info.to_dict() == expected


def test_load_normalization():
    proj = config.read_pyproject_file(data_dir / "normalization" / "pyproject.toml")
    assert proj.metadata["name"] == "my-python-module"


def test_load_pep621():
    proj = config.read_pyproject_file(data_dir / "pep621" / "pyproject.toml")
    assert proj.metadata["name"] == "module1"
    assert proj.metadata["version"] == "1.0.0"
    assert proj.metadata["description_content_type"] == "text/x-rst"
    # Remove all whitespace from requirements so we don't check exact format:
    assert {r.replace(" ", "") for r in proj.metadata["requires_dist"]} == {
        "docutils",
        "requests>=2.18",
        'pytest;extra=="test"',  # from [project.optional-dependencies]
        "mock;extra==\"test\"and(python_version<'3.6')",
    }
    assert proj.metadata["author_email"] == "Sir Röbin <robin@camelot.uk>"
    assert proj.entrypoints["flit_test_example"]["foo"] == "module1:main"


def test_load_pep621_nodynamic():
    proj = config.read_pyproject_file(data_dir / "pep621_nodynamic" / "pyproject.toml")
    assert proj.metadata["name"] == "module1"
    assert proj.metadata["version"] == "0.3"
    assert proj.metadata["summary"] == "Statically specified description"


def test_misspelled_key():
    with pytest.raises(config.ConfigError) as e_project_info:
        config.read_pyproject_file(data_dir / "misspelled-key.toml")

    assert "Description file" in str(e_project_info.value)


def test_description_file():
    project_info = config.read_pyproject_file(
        data_dir / "description-file" / "package1.toml"
    )
    assert project_info.metadata["summary"] == "short description"
    assert project_info.metadata["description"] == "Sample description for test.\n"
    assert project_info.metadata["description_content_type"] == "text/x-rst"


def test_missing_description_file():
    with pytest.raises(config.ConfigError, match=r"Description file .* does not exist"):
        config.read_pyproject_file(data_dir / "description-missing" / "package1.toml")


def test_bad_description_extension(caplog):
    project_info = config.read_pyproject_file(
        data_dir / "description-bad" / "bad-description-ext.toml"
    )
    assert project_info.metadata["description_content_type"] is None
    assert any(
        (r.levelno == logging.WARN and "Unknown extension" in r.msg)
        for r in caplog.records
    )


def test_with_wheel_path_prefixes_to_strip():
    project_info = config.read_pyproject_file(
        data_dir / "advanced_toml" / "pyproject.toml"
    )
    expected = {
        "entrypoints": {},
        "excludes": ["**/.git/*", "**/.hg/*"],
        "includes": ["src/module3.py"],
        "metadata": {
            "name": "module3",
            "requires_dist": [],
            "summary": "module3 desc",
            "version": "1.0",
        },
        "metadata_files": ["LICENSE"],
        "referenced_files": [],
        "reqs_by_extra": {},
        "sdist_extra_excludes": ["**/.git/*", "**/.hg/*"],
        "sdist_extra_includes": ["foo.txt"],
        "editable_paths": ["src"],
        "wheel_path_prefixes_to_strip": ["src"],
    }
    assert project_info.to_dict() == expected


path_include_tests = [
    ("../bar", "contains relative .."),
    ("foo/../../bar", "contains relative .."),
    ("/home", "absolute path"),
    ("foo:bar", "bad character"),
]


@pytest.mark.parametrize(("path", "err_match"), path_include_tests)
def test_bad_includes_paths(path, err_match):
    toml_cfg = {
        "project": {
            "name": "foo",
            "version": "1.0.0",
            "description": "desc",
        },
        "tool": {"flot": {"includes": [path]}},
    }
    with pytest.raises(config.ConfigError, match=err_match):
        config.prep_pyproject_config(toml_cfg, None)


@pytest.mark.parametrize(("path", "err_match"), path_include_tests)
def test_bad_sdist_extra_includes_paths(path, err_match):
    toml_cfg = {
        "project": {
            "name": "foo",
            "version": "1.0.0",
            "description": "desc",
        },
        "tool": {"flot": {"includes": ["foo"], "sdist_extra_includes": [path]}},
    }

    with pytest.raises(config.ConfigError, match=err_match):
        config.prep_pyproject_config(toml_cfg, None)


@pytest.mark.parametrize(("path", "err_match"), path_include_tests)
def test_bad_metadata_files_paths(path, err_match):
    toml_cfg = {
        "project": {
            "name": "foo",
            "version": "1.0.0",
            "description": "desc",
        },
        "tool": {"flot": {"includes": ["foo"], "metadata_files": [path]}},
    }

    with pytest.raises(config.ConfigError, match=err_match):
        config.prep_pyproject_config(toml_cfg, None)


@pytest.mark.parametrize(
    ("proj_bad", "err_match"),
    [
        ({"version": 1}, r"\bstr\b"),
        ({"license": {"fromage": 2}}, "[Uu]nrecognised"),
        ({"license": {"file": "LICENSE", "text": "xyz"}}, "both"),
        ({"license": {}}, "required"),
        ({"keywords": "foo"}, "list"),
        ({"keywords": ["foo", 7]}, "strings"),
        ({"entry-points": {"foo": "module1:main"}}, "entry-point.*tables"),
        ({"entry-points": {"group": {"foo": 7}}}, "entry-point.*string"),
        (
            {"entry-points": {"gui_scripts": {"foo": "a:b"}}},
            r"\[project\.gui-scripts\]",
        ),
        ({"scripts": {"foo": 7}}, "scripts.*string"),
        ({"gui-scripts": {"foo": 7}}, "gui-scripts.*string"),
        ({"optional-dependencies": {"test": "requests"}}, "list.*optional-dep"),
        ({"optional-dependencies": {"test": [7]}}, "string.*optional-dep"),
        ({"dynamic": ["classifiers"]}, "does not support dynamic"),
        ({"dynamic": ["version"]}, "does not support dynamic"),
        ({"authors": ["thomas"]}, r"author.*\bdict"),
        ({"maintainers": [{"title": "Dr"}]}, r"maintainer.*title"),
    ],
)
def test_bad_pep621_project_info(proj_bad, err_match):
    proj = {"name": "module1", "version": "1.0", "description": "x"}
    proj.update(proj_bad)
    with pytest.raises(config.ConfigError, match=err_match):
        config.read_pep621_metadata(proj, data_dir / "pep621")


@pytest.mark.parametrize(
    ("readme", "err_match"),
    [
        ({"file": "README.rst"}, "required"),
        ({"file": "README.rst", "content-type": "text/x-python"}, "content-type"),
        ("/opt/README.rst", "relative"),
        ({"file": "README.rst", "text": "", "content-type": "text/x-rst"}, "both"),
        ({"content-type": "text/x-rst"}, "required"),
        (
            {"file": "README.rst", "content-type": "text/x-rst", "a": "b"},
            "[Uu]nrecognised",
        ),
        (5, r"readme.*string"),
    ],
)
def test_bad_pep621_readme(readme, err_match):
    proj = {"name": "module1", "version": "1.0", "description": "x", "readme": readme}
    with pytest.raises(config.ConfigError, match=err_match):
        config.read_pep621_metadata(proj, data_dir / "pep621")



#
# SPDX-License-Identifier: BSD-2-clause AND BSD-3-clause
# Copyright (c) 2015, Thomas Kluyver and contributors
# Copyright (c) Donald Stufft and individual contributors.
# Based on https://github.com/pypa/flit/ and heavily modified
# Regex from packaging, via PEP 440

"""Normalize version number according to PEP 440"""
import logging
import os
import re

log = logging.getLogger(__name__)

VERSION_PERMISSIVE = re.compile(
    r"""
    \s*v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_\.]?
                (?P<post_l>post|rev|r)
                [-_\.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_\.]?
            (?P<dev_l>dev)
            [-_\.]?
            (?P<dev_n>[0-9]+)?
        )?
    )
    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
\s*$""",
    re.VERBOSE,
)

pre_spellings = {
    "a": "a",
    "alpha": "a",
    "b": "b",
    "beta": "b",
    "rc": "rc",
    "c": "rc",
    "pre": "rc",
    "preview": "rc",
}


def normalize_version(orig_version):
    """Normalise version number according to rules in PEP 440

    Raises InvalidVersion if the version does not match PEP 440. This can be
    overridden with the FLIT_ALLOW_INVALID environment variable.

    https://www.python.org/dev/peps/pep-0440/#normalization
    """
    version = orig_version.lower()
    m = VERSION_PERMISSIVE.match(version)
    if not m:
        if os.environ.get("FLIT_ALLOW_INVALID"):
            log.warning(
                "Invalid version number {!r} allowed by FLIT_ALLOW_INVALID".format(
                    orig_version
                )
            )
            return version
        else:
            from .common import InvalidVersion

            raise InvalidVersion(
                "Version number {!r} does not match PEP 440 rules".format(orig_version)
            )

    components = []
    add = components.append

    epoch, release = m.group("epoch", "release")
    if epoch is not None:
        add(str(int(epoch)) + "!")
    add(".".join(str(int(rp)) for rp in release.split(".")))

    pre_l, pre_n = m.group("pre_l", "pre_n")
    if pre_l is not None:
        pre_l = pre_spellings[pre_l]
        pre_n = "0" if pre_n is None else str(int(pre_n))
        add(pre_l + pre_n)

    post_n1, post_l, post_n2 = m.group("post_n1", "post_l", "post_n2")
    if post_n1 is not None:
        add(".post" + str(int(post_n1)))
    elif post_l is not None:
        post_n = "0" if post_n2 is None else str(int(post_n2))
        add(".post" + str(int(post_n)))

    dev_l, dev_n = m.group("dev_l", "dev_n")
    if dev_l is not None:
        dev_n = "0" if dev_n is None else str(int(dev_n))
        add(".dev" + dev_n)

    local = m.group("local")
    if local is not None:
        local = local.replace("-", ".").replace("_", ".")
        l = [str(int(c)) if c.isdigit() else c for c in local.split(".")]
        add("+" + ".".join(l))

    version = "".join(components)
    if version != orig_version:
        log.warning(
            "Version number normalised: {!r} -> {!r} (see PEP 440)".format(
                orig_version, version
            )
        )
    return version



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

from pathlib import Path

import pytest

from flot.common import FilePatterns
from flot.common import InvalidVersion
from flot.common import NoVersionError
from flot.common import check_version
from flot.common import iter_files
from flot.common import normalize_file_permissions

data_dir = Path(__file__).parent / "data"


def test_version_raise():
    with pytest.raises(InvalidVersion):
        check_version("a.1.0.beta0")

    with pytest.raises(InvalidVersion):
        check_version("3!")

    with pytest.raises(InvalidVersion):
        check_version((1, 2))

    with pytest.raises(NoVersionError):
        check_version(None)

    assert check_version("4.1.0beta1") == "4.1.0b1"
    assert check_version("v1.2") == "1.2"


def test_normalize_file_permissions():
    assert normalize_file_permissions(0o100664) == 0o100644  # regular file
    assert normalize_file_permissions(0o40775) == 0o40755  # directory


def test_iter_files_can_iter_dir(get_test_data):
    test_dir = get_test_data("module3")
    expected = ["LICENSE", "pyproject.toml", "src/module3.py"]
    results = [str(Path(f).relative_to(test_dir)) for f in iter_files(test_dir)]
    assert results == expected


def test_iter_files_can_iter_file(get_test_data):
    test_dir = get_test_data("module3")
    expected = ["LICENSE", "pyproject.toml", "src/module3.py"]
    results = [str(Path(f).relative_to(test_dir)) for f in iter_files(test_dir)]
    assert results == expected


def test_iter_files_ignores_pycache_and_pyc(copy_test_data):
    test_dir = Path(copy_test_data("module3"))

    pyc1 = test_dir / "python1.pyc"
    with pyc1.open("w") as f:
        f.write("")

    pycache = test_dir / "__pycache__"
    pycache.mkdir()

    pyc2 = pycache / "python2.pyc"
    with pyc2.open("w") as f:
        f.write("")

    expected = ["LICENSE", "pyproject.toml", "src/module3.py"]
    results = [str(Path(f).relative_to(test_dir)) for f in iter_files(test_dir)]
    assert results == expected


def test_apply_includes_excludes_with_simple_file_include(get_test_data):
    base_dir = get_test_data("module3")
    files = list(iter_files(base_dir))
    rel_files = [str(Path(f).relative_to(base_dir)) for f in files]
    assert rel_files == ["LICENSE", "pyproject.toml", "src/module3.py"]

    include_patterns = FilePatterns(patterns=["pyproject.toml"], base_dir=base_dir)
    assert list(map(str, include_patterns.files)) == ["pyproject.toml"]

    results = [
        str(p)
        for p in FilePatterns.apply_includes_excludes(
            base_dir=base_dir,
            files=files,
            include_patterns=include_patterns,
            relative_paths=True,
        )
    ]
    expected = ["pyproject.toml"]
    assert results == expected


def test_apply_includes_excludes_with_file_includes_and_excludes(get_test_data):
    base_dir = get_test_data("module4")
    files = list(iter_files(base_dir))
    rel_files = [str(Path(f).relative_to(base_dir)) for f in files]
    assert rel_files == [
        "LICENSE",
        "pyproject.toml",
        "src/module3.py",
        "src/module4.py",
    ]

    include_patterns = FilePatterns(
        patterns=["pyproject.toml", "src/**/*"], base_dir=base_dir
    )
    expected = {"pyproject.toml", "src/module3.py", "src/module4.py"}
    assert include_patterns.files == expected

    exclude_patterns = FilePatterns(patterns=["src/module3.py"], base_dir=base_dir)
    assert exclude_patterns.files == {"src/module3.py"}

    results = [
        str(p)
        for p in FilePatterns.apply_includes_excludes(
            base_dir=base_dir,
            files=files,
            include_patterns=include_patterns,
            exclude_patterns=exclude_patterns,
            relative_paths=True,
        )
    ]
    expected = ["pyproject.toml", "src/module4.py"]
    assert results == expected


def test_apply_includes_excludes_with_nested_file_includes_and_excludes(get_test_data):
    base_dir = get_test_data("module5")
    files = list(iter_files(base_dir))
    rel_files = [str(Path(f).relative_to(base_dir)) for f in files]
    assert rel_files == [
        "LICENSE",
        "pyproject.toml",
        "src/deep/nested/foo.py",
        "src/module3.py",
        "src/module4.py",
    ]

    include_patterns = FilePatterns(
        patterns=["pyproject.toml", "src/**/*"], base_dir=base_dir
    )
    expected = {
        "src/module4.py",
        "pyproject.toml",
        "src/deep/nested/foo.py",
        "src/module3.py",
    }
    assert include_patterns.files == expected

    exclude_patterns = FilePatterns(
        patterns=["src/module3.py", "src/deep/**/*"], base_dir=base_dir
    )
    assert exclude_patterns.files == {"src/module3.py", "src/deep/nested/foo.py"}

    results = [
        str(p)
        for p in FilePatterns.apply_includes_excludes(
            base_dir=base_dir,
            files=files,
            include_patterns=include_patterns,
            exclude_patterns=exclude_patterns,
            relative_paths=True,
        )
    ]
    expected = ["pyproject.toml", "src/module4.py"]
    assert results == expected



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

import hashlib
import logging
import re
from pathlib import Path

log = logging.getLogger(__name__)

from .versionno import normalize_version


class FilePatterns:
    """
    Compute a set of file Path inclusion or exclusion patterns relative to
    base_dir
    """

    def __init__(self, patterns, base_dir):
        self.base_dir = Path(base_dir).absolute()
        self.files = set()

        log.debug(f"FilePatterns: patterns: {patterns} base_dir: {base_dir}")

        for pattern in patterns:
            # Always remove trailing / as Path.rglob will return only dirs in
            # Python 3.11 and up with this
            pattern = pattern.rstrip("/")
            for path in sorted(self.base_dir.glob(pattern)):
                if path.is_dir():
                    continue
                rel_path = str(path.relative_to(base_dir))
                log.debug(f"FilePatterns.add: {rel_path}")
                self.files.add(rel_path)

    def match_file(self, rel_path):
        log.debug(
            f"match_file: {rel_path} is matched: {rel_path in self.files} in {self.files!r}"
        )
        return str(rel_path) in self.files

    @classmethod
    def apply_includes_excludes(
        cls,
        base_dir,
        files,
        include_patterns,
        exclude_patterns=None,
        relative_paths=False,
    ):
        """
        Return a sorted list of file Path objects filtered from
        applying include and exclude patterns.
        """
        if not exclude_patterns:
            exclude_patterns = cls([], base_dir=base_dir)

        base_dir = Path(base_dir)
        all_files = [Path(f).relative_to(base_dir) for f in files]
        paths = set()
        for path in all_files:
            if include_patterns.match_file(path) and not exclude_patterns.match_file(
                path
            ):
                if not relative_paths:
                    path = base_dir / path
                paths.add(path)
        return sorted(paths)


class FileSelector:
    """A mixin for builders to help select files"""

    def select_files(self):
        """
        Yield tuple of (absolute Path, relative Path) for files to include in
        all distributions (wheel and sdist).
        """
        include_patterns = self.include_patterns
        exclude_patterns = self.exclude_patterns
        return self._select_files(self.base_dir, include_patterns, exclude_patterns)

    def select_extra_sdist_files(self):
        """
        Yield tuple of (absolute Path, relative Path) for files to include as
        extra in the sdist distribution.
        """
        include_patterns = self.sdist_extra_include_patterns
        exclude_patterns = self.sdist_extra_exclude_patterns
        return self._select_files(self.base_dir, include_patterns, exclude_patterns)

    def select_metafiles(self):
        """
        Yield tuple of (absolute Path, relative Path) for files to include in
        a wheel metadata. Also include in an sdist.
        """
        include_patterns = self.metadata_file_patterns
        return self._select_files(self.base_dir, include_patterns)

    @classmethod
    def _select_files(cls, base_dir, include_patterns, exclude_patterns=None):
        """
        Yield tuple of (absolute Path, relative Path) for files to include as
        extra in the sdist distribution.
        """
        log.debug(f"_select_files: project_dir: {base_dir}")
        files = iter_files(base_dir)
        for abs_path in FilePatterns.apply_includes_excludes(
            base_dir=base_dir,
            files=files,
            include_patterns=include_patterns,
            exclude_patterns=exclude_patterns,
        ):
            yield abs_path, abs_path.relative_to(base_dir)


def iter_files(base_dir: Path):
    """Iterate over the files contained in this base_dir Path.

    Yields absolute paths - caller may want to make them relative.
    Excludes any __pycache__ and *.pyc files.
    """
    base_dir = Path(base_dir)
    if base_dir.is_file():
        yield base_dir
    else:
        yield from (
            p
            for p in sorted(base_dir.rglob("*"))
            if p.is_file() and p.name != "__pycache__" and p.suffix != ".pyc"
        )


class NoVersionError(ValueError):
    pass


class InvalidVersion(ValueError):
    pass


def check_version(version):
    """
    Check whether a given version string match PEP 440, and do normalisation.

    Raise InvalidVersion/NoVersionError with relevant information if
    version is invalid.

    Log a warning if the version is not canonical with respect to PEP 440.

    Returns the version in canonical PEP 440 format.
    """
    if not version:
        raise NoVersionError(
            "Cannot package module without a version string. "
            'Please define a `__version__ = "x.y.z"` in your module.'
        )
    if not isinstance(version, str):
        raise InvalidVersion(
            "__version__ must be a string, not {}.".format(type(version))
        )

    # Import here to avoid circular import
    version = normalize_version(version)

    return version


def parse_entry_point(ep):
    """Check and parse a 'package.module:func' style entry point specification.

    Returns (modulename, funcname)
    """
    if ":" not in ep:
        raise ValueError("Invalid entry point (no ':'): %r" % ep)
    mod, func = ep.split(":")

    for piece in func.split("."):
        if not piece.isidentifier():
            raise ValueError("Invalid entry point: %r is not an identifier" % piece)
    for piece in mod.split("."):
        if not piece.isidentifier():
            raise ValueError("Invalid entry point: %r is not a module path" % piece)

    return mod, func


def write_entry_points(d, fp):
    """Write entry_points.txt from a two-level dict

    Sorts on keys to ensure results are reproducible.
    """
    for group_name in sorted(d):
        fp.write("[{}]\n".format(group_name))
        group = d[group_name]
        for name in sorted(group):
            val = group[name]
            fp.write("{}={}\n".format(name, val))
        fp.write("\n")


def hash_file(path, algorithm="sha256"):
    with open(path, "rb") as f:
        h = hashlib.new(algorithm, f.read())
    return h.hexdigest()


def normalize_file_permissions(st_mode):
    """Normalize the permission bits in the st_mode field from stat to 644/755

    Popular VCSs only track whether a file is executable or not. The exact
    permissions can vary on systems with different umasks. Normalising
    to 644 (non executable) or 755 (executable) makes builds more reproducible.
    """
    # Set 644 permissions, leaving higher bits of st_mode unchanged
    new_mode = (st_mode | 0o644) & ~0o133
    if st_mode & 0o100:
        new_mode |= 0o111  # Executable: 644 -> 755
    return new_mode


class Metadata:
    summary = None
    home_page = None
    author = None
    author_email = None
    maintainer = None
    maintainer_email = None
    license = None
    description = None
    keywords = None
    download_url = None
    requires_python = None
    description_content_type = None

    platform = ()
    supported_platform = ()
    classifiers = ()
    requires = ()
    obsoletes = ()
    project_urls = ()
    provides_dist = ()
    requires_dist = ()
    obsoletes_dist = ()
    requires_external = ()
    provides_extra = ()

    metadata_version = "2.1"

    def __init__(self, data):
        data = data.copy()
        self.name = data.pop("name")
        self.version = data.pop("version")

        for k, v in data.items():
            assert hasattr(self, k), "data does not have attribute '{}'".format(k)
            setattr(self, k, v)

    def _normalise_name(self, n):
        return n.lower().replace("-", "_")

    def write_metadata_file(self, fp):
        """Write out metadata in the email headers format"""
        fields = [
            "Metadata-Version",
            "Name",
            "Version",
        ]
        optional_fields = [
            "Summary",
            "Home-page",
            "License",
            "Keywords",
            "Author",
            "Author-email",
            "Maintainer",
            "Maintainer-email",
            "Requires-Python",
            "Description-Content-Type",
        ]

        for field in fields:
            value = getattr(self, self._normalise_name(field))
            fp.write("{}: {}\n".format(field, value))

        for field in optional_fields:
            value = getattr(self, self._normalise_name(field))
            if value is not None:
                # TODO: verify which fields can be multiline
                # The spec has multiline examples for Author, Maintainer &
                # License (& Description, but we put that in the body)
                # Indent following lines with 8 spaces:
                value = "\n        ".join(value.splitlines())
                fp.write("{}: {}\n".format(field, value))

        for clsfr in self.classifiers:
            fp.write("Classifier: {}\n".format(clsfr))

        for req in self.requires_dist:
            fp.write("Requires-Dist: {}\n".format(req))

        for url in self.project_urls:
            fp.write("Project-URL: {}\n".format(url))

        for extra in self.provides_extra:
            fp.write("Provides-Extra: {}\n".format(extra))

        if self.description is not None:
            fp.write("\n" + self.description + "\n")


def normalize_dist_name(name: str, version: str) -> str:
    """Normalizes a name and a PEP 440 version

    The resulting string is valid as dist-info folder name
    and as first part of a wheel filename

    See https://packaging.python.org/specifications/binary-distribution-format/#escaping-and-unicode
    """
    normalized_name = re.sub(r"[-_.]+", "_", name, flags=re.UNICODE).lower()
    assert check_version(version) == version
    assert "-" not in version, "Normalized versions can’t have dashes"
    return "{}-{}".format(normalized_name, version)


def dist_info_name(distribution, version):
    """Get the correct name of the .dist-info dir"""
    return normalize_dist_name(distribution, version) + ".dist-info"



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

import shutil
import tarfile
import tempfile
from io import BytesIO
from pathlib import Path

from testpath import assert_isfile

from flot import sdist

data_dir = Path(__file__).parent / "data"


def unpack(path):
    t = tempfile.TemporaryDirectory()
    shutil.unpack_archive(str(path), extract_dir=t.name)
    return t


def test_make_sdist(tmp_path):
    # Smoke test of making a complete sdist
    builder = sdist.SdistBuilder.from_pyproject_file(
        data_dir / "package1" / "pyproject.toml"
    )
    sd = builder.build(tmp_path)
    assert_isfile(tmp_path / "package1-0.0.1.tar.gz")

    with unpack(sd) as unpacked:
        files = sorted(
            str(p.relative_to(unpacked))
            for p in Path(unpacked).glob("**/*")
            if p.is_file()
        )
        assert files == [
            "package1-0.0.1/PKG-INFO",
            "package1-0.0.1/__init__.py",
            "package1-0.0.1/data_dir/foo.sh",
            "package1-0.0.1/foo.py",
            "package1-0.0.1/package1/__init__.py",
            "package1-0.0.1/package1/data_dir/foo.sh",
            "package1-0.0.1/package1/foo.py",
            "package1-0.0.1/package1/subpkg/__init__.py",
            "package1-0.0.1/package1/subpkg/sp_data_dir/test.json",
            "package1-0.0.1/package1/subpkg2/__init__.py",
            "package1-0.0.1/pyproject.toml",
            "package1-0.0.1/subpkg/__init__.py",
            "package1-0.0.1/subpkg/sp_data_dir/test.json",
            "package1-0.0.1/subpkg2/__init__.py",
        ]


def test_make_sdist_with_extra_includes_and_ignored_wheel_suffix_strips(tmp_path):
    # Smoke test of making a complete sdist
    builder = sdist.SdistBuilder.from_pyproject_file(
        data_dir / "module5" / "pyproject.toml"
    )
    sd = builder.build(tmp_path)

    with unpack(sd) as unpacked:
        files = sorted(
            str(p.relative_to(unpacked))
            for p in Path(unpacked).glob("**/*")
            if p.is_file()
        )
        assert files == [
            "module3-1.0/LICENSE",
            "module3-1.0/PKG-INFO",
            "module3-1.0/pyproject.toml",
            "module3-1.0/src/module3.py",
            "module3-1.0/src/module4.py",
        ]


def test_make_sdist_with_extra_includes_and_ignored_wheel_suffix_strips_and_rename_pyproject(
    tmp_path,
):
    builder = sdist.SdistBuilder.from_pyproject_file(
        data_dir / "module6" / "not-py-project.foo"
    )
    sd = builder.build(tmp_path)

    with unpack(sd) as unpacked:
        files = sorted(
            str(p.relative_to(unpacked))
            for p in Path(unpacked).glob("**/*")
            if p.is_file()
        )
        assert files == [
            "module3-1.0/LICENSE",
            "module3-1.0/PKG-INFO",
            "module3-1.0/extras/some.txt",
            "module3-1.0/pyproject.toml",
            "module3-1.0/src/deep/nested/foo.py",
            "module3-1.0/src/module3.py",
            "module3-1.0/src/module4.py",
        ]


def test_make_sdist_pep621(tmp_path):
    builder = sdist.SdistBuilder.from_pyproject_file(
        data_dir / "pep621" / "pyproject.toml"
    )
    path = builder.build(tmp_path)
    assert path == tmp_path / "module1-1.0.0.tar.gz"
    assert_isfile(path)


def test_make_sdist_pep621_nodynamic(tmp_path):
    builder = sdist.SdistBuilder.from_pyproject_file(
        data_dir / "pep621_nodynamic" / "pyproject.toml"
    )
    path = builder.build(tmp_path)
    assert path == tmp_path / "module1-0.3.tar.gz"
    assert_isfile(path)


def test_clean_tarinfo():
    with tarfile.open(mode="w", fileobj=BytesIO()) as tf:
        ti = tf.gettarinfo(str(data_dir / "module1.py"))
    cleaned = sdist.clean_tarinfo(ti, mtime=42)
    assert cleaned.uid == 0
    assert cleaned.uname == ""
    assert cleaned.mtime == 42


def test_include_exclude():
    builder = sdist.SdistBuilder.from_pyproject_file(
        data_dir / "inclusion" / "pyproject.toml"
    )
    files = sorted(str(rel) for _abs, rel in builder.select_all_files())
    expected = [
        "LICENSES/README",
        "doc/subdir/test.txt",
        "doc/test.rst",
        "module1.py",
        "pyproject.toml",
    ]
    assert files == expected


def test_data_dir():
    builder = sdist.SdistBuilder.from_pyproject_file(
        data_dir / "with_data_dir" / "pyproject.toml"
    )
    files = sorted(str(rel) for _abs, rel in builder.select_all_files())
    expected = ["LICENSE", "data/share/man/man1/foo.1", "pyproject.toml"]
    assert files == expected


def test_pep625(tmp_path):
    builder = sdist.SdistBuilder.from_pyproject_file(
        data_dir / "normalization" / "pyproject.toml"
    )
    path = builder.build(tmp_path)
    assert path == tmp_path / "my_python_module-0.0.1.tar.gz"
    assert_isfile(path)



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

import contextlib
import hashlib
import io
import logging
import os
import stat
from base64 import urlsafe_b64encode
from datetime import datetime
from pathlib import Path
from typing import Optional
from zipfile import ZIP_DEFLATED
from zipfile import ZipFile
from zipfile import ZipInfo

from . import common

log = logging.getLogger(__name__)


def _write_WHEEL_file(f, wheel_tag=None):
    from flot import __version__

    wheel_tag = wheel_tag or "py3-none-any"
    is_pure_lib = "true" if wheel_tag == "py3-none-any" else "false"

    wheel_file_template = f"""\
Wheel-Version: 1.0
Generator: flot {__version__}
Root-Is-Purelib: {is_pure_lib}
Tag: {wheel_tag}
"""
    f.write(wheel_file_template)


def _set_zinfo_mode(zinfo, mode):
    # Set the bits for the mode
    zinfo.external_attr = mode << 16


def zip_timestamp_from_env() -> Optional[tuple]:
    """Prepare a timestamp from $SOURCE_DATE_EPOCH, if set"""
    try:
        # If SOURCE_DATE_EPOCH is set (e.g. by Debian), it's used for
        # timestamps inside the zip file.
        d = datetime.utcfromtimestamp(int(os.environ["SOURCE_DATE_EPOCH"]))
    except (KeyError, ValueError):
        # Otherwise, we'll use the mtime of files, and generated files will
        # default to 2016-1-1 00:00:00
        return None

    if d.year >= 1980:
        log.info("Zip timestamps will be from SOURCE_DATE_EPOCH: %s", d)
        # zipfile expects a 6-tuple, not a datetime object
        return d.year, d.month, d.day, d.hour, d.minute, d.second
    else:
        log.info("SOURCE_DATE_EPOCH is below the minimum for zip file timestamps")
        log.info("Zip timestamps will be 1980-01-01 00:00:00")
        return 1980, 1, 1, 0, 0, 0


class WheelBuilder(common.FileSelector):
    def __init__(
        self,
        pyproject_file,
        metadata,
        entrypoints,
        includes=(),
        excludes=(),
        path_prefixes_to_strip=(),
        editable_paths=(),
        metadata_files=(),
        wheel_tag=None,
    ):
        """
        Build a wheel for the files include/exclude patterns provided
        """
        log.debug(f"WheelBuilder: {pyproject_file}")
        self.pyproject_file = Path(pyproject_file).absolute()
        self.base_dir = self.pyproject_file.parent
        self.metadata = metadata
        self.entrypoints = entrypoints

        self.include_patterns = common.FilePatterns(
            patterns=includes, base_dir=self.base_dir
        )
        self.exclude_patterns = common.FilePatterns(
            patterns=excludes, base_dir=self.base_dir
        )
        self.path_prefixes_to_strip = path_prefixes_to_strip

        if not editable_paths:
            editable_paths = ["."]

        self.editable_paths = editable_paths or []

        # this is the only implict addition that is too easily overseen
        if not metadata_files:
            metadata_files = (
                "*LICEN[SC]E*",
                "*COPYING*",
                "*COPYIRIGHT*",
            )
        self.metadata_file_patterns = common.FilePatterns(
            patterns=metadata_files, base_dir=self.base_dir
        )

        self.records = []
        self.source_time_stamp = zip_timestamp_from_env()

        self.wheel_tag = wheel_tag or "py3-none-any"
        self.wheel_zip = None

    @classmethod
    def from_pyproject_file(
        cls,
        pyproject_file=None,
        wheel_tag=None,
    ):
        """
        Create a wheel in target_dir, default: <pyproject file parent>/dist
        from an arbitrary pyproject.toml file path, default: "pyproject.toml".
        Use an optional wheel tag, default: "py3-none-any"
        """
        pyproject_file = pyproject_file or "pyproject.toml"
        wheel_tag = wheel_tag or "py3-none-any"

        from .config import read_pyproject_file

        pyproject_file = Path(pyproject_file).absolute()
        project_info = read_pyproject_file(pyproject_file)

        return cls(
            pyproject_file=pyproject_file,
            metadata=common.Metadata(project_info.metadata),
            entrypoints=project_info.entrypoints,
            includes=project_info.includes,
            excludes=project_info.excludes,
            path_prefixes_to_strip=project_info.wheel_path_prefixes_to_strip,
            editable_paths=project_info.editable_paths,
            metadata_files=project_info.metadata_files,
            wheel_tag=wheel_tag,
        )

    @property
    def wheel_filename(self):
        dist_name = common.normalize_dist_name(
            self.metadata.name, self.metadata.version
        )
        return f"{dist_name}-{self.wheel_tag}.whl"

    def build(self, output_dir=None, editable=False):
        """
        Build wheel and return wheel file path.
        """
        output_dir = (
            output_dir and Path(output_dir) or Path(self.pyproject_file.parent) / "dist"
        )
        output_dir.mkdir(parents=True, exist_ok=True)

        target = output_dir / self.wheel_filename

        with ZipFile(target, "w", compression=ZIP_DEFLATED) as self.wheel_zip:
            if editable:
                self.add_editable_pth_file()
            else:
                self.add_files()
            self.write_metadata()
            self.write_record()
        return target

    def add_files(self):
        log.info("Copying selected file(s) to wheel")
        prefixes = self.path_prefixes_to_strip
        for abs_path, rel_path in self.select_files():
            rel_path = strip_prefixes(rel_path, prefixes)
            self._add_file(abs_path, rel_path)

    def add_editable_pth_file(self):
        """
        Write a .pth path file in an editable wheel using configured editable_paths.
        """
        base_dir = self.base_dir.absolute()
        with self._write_to_zip(self.metadata.name + ".pth") as pth_file:
            # log.debug(f"writing editable_path: {base_dir}")
            # pth_file.write(str(base_dir) + "\n")
            for rel_path in self.editable_paths:
                log.debug(f"writing editable_path: {base_dir / rel_path}")
                pth_file.write(str(base_dir / rel_path) + "\n")

    @property
    def dist_info_dir(self):
        return common.dist_info_name(self.metadata.name, self.metadata.version)

    def _add_file(self, abs_path, rel_path):
        log.debug(f"Adding {abs_path!r} to zip file as {rel_path!r}")
        abs_path, rel_path = str(abs_path), str(rel_path)
        if os.sep != "/":
            # We always want to have /-separated paths in the zip file and in
            # RECORD
            rel_path = rel_path.replace(os.sep, "/")

        zinfo = ZipInfo.from_file(filename=abs_path, arcname=rel_path)
        if self.source_time_stamp:
            # Set timestamps in zipfile for reproducible build
            zinfo.date_time = self.source_time_stamp

        # Normalize permission bits to either 755 (executable) or 644
        st_mode = os.stat(abs_path).st_mode
        new_mode = common.normalize_file_permissions(st_mode)
        _set_zinfo_mode(zinfo, new_mode & 0xFFFF)  # Unix attributes

        if stat.S_ISDIR(st_mode):
            zinfo.external_attr |= 0x10  # MS-DOS directory flag

        zinfo.compress_type = ZIP_DEFLATED

        hashsum = hashlib.sha256()
        with open(abs_path, "rb") as src, self.wheel_zip.open(zinfo, "w") as dst:
            while True:
                buf = src.read(1024 * 8)
                if not buf:
                    break
                hashsum.update(buf)
                dst.write(buf)

        # keep for RECORDS
        size = os.stat(abs_path).st_size
        hash_digest = urlsafe_b64encode(hashsum.digest()).decode("ascii").rstrip("=")
        self.records.append((rel_path, hash_digest, size))

    @contextlib.contextmanager
    def _write_to_zip(self, rel_path, mode=0o644):
        sio = io.StringIO()
        yield sio

        log.debug("Writing data to %s in zip file", rel_path)
        # The default is a fixed timestamp rather than the current time, so
        # that building a wheel twice on the same computer can automatically
        # give you the exact same result.
        date_time = self.source_time_stamp or (2016, 1, 1, 0, 0, 0)
        zi = ZipInfo(rel_path, date_time)
        # Also sets bit 0x8000 for "regular file" (S_IFREG)
        _set_zinfo_mode(zi, mode | stat.S_IFREG)
        b = sio.getvalue().encode("utf-8")
        hashsum = hashlib.sha256(b)
        hash_digest = urlsafe_b64encode(hashsum.digest()).decode("ascii").rstrip("=")
        self.wheel_zip.writestr(zi, b, compress_type=ZIP_DEFLATED)
        self.records.append((rel_path, hash_digest, len(b)))

    def write_metadata(self):
        log.info("Writing metadata files")

        if self.entrypoints:
            with self._write_to_zip(self.dist_info_dir + "/entry_points.txt") as f:
                common.write_entry_points(self.entrypoints, f)

        with self._write_to_zip(self.dist_info_dir + "/WHEEL") as f:
            _write_WHEEL_file(f, wheel_tag=self.wheel_tag)

        with self._write_to_zip(self.dist_info_dir + "/METADATA") as f:
            self.metadata.write_metadata_file(f)

        # add all metadata files as-is to the root of the wheel dist-info dir
        for abs_meta, rel_meta in self.select_metafiles():
            # metadata files go directly in the dist_info directory, not in subdirs
            self._add_file(abs_meta, f"{self.dist_info_dir}/{rel_meta.name}")

    def write_record(self):
        log.info("Writing the record of files")
        # Write a record of the files in the wheel
        with self._write_to_zip(self.dist_info_dir + "/RECORD") as f:
            for path, hsh, size in self.records:
                f.write("{},sha256={},{}\n".format(path, hsh, size))
            # RECORD itself is recorded with no hash or size
            f.write(self.dist_info_dir + "/RECORD,,\n")


def strip_prefixes(rel_path, prefixes):
    """
    Return a ``rel_path`` relative path  string stripped from the first found
    prefix of a list of ``prefixes`` path strings where rel_path starts with
    this prefix.
    Return the path unchanged if it has is no matching prefix.

    For example::
    >>> assert strip_prefixes("src/foo.py", ["lib", "src"]) == "foo.py"
    True
    >>> assert strip_prefixes("src/lib/foo.py", ["lib", "src"]) == "lib/foo.py"
    True
    >>> assert strip_prefixes("src/lib/foo.py", ["lib"]) == "src/lib/foo.py"
    True
    >>> assert strip_prefixes("foo.py", ["lib"]) == "foo.py"
    True
    >>> assert strip_prefixes("foo.py", []) == "foo.py"
    True
    """
    rel_path = str(rel_path)
    for prefix in prefixes:
        if rel_path.startswith(prefix):
            return rel_path.replace(prefix, "", 1).lstrip("/")
    return rel_path


def make_wheel(
    pyproject_file=None,
    output_dir=None,
    wheel_tag=None,
    editable=False,
):
    """
    Make a wheel, return the path.
    """
    builder = WheelBuilder.from_pyproject_file(
        pyproject_file=pyproject_file,
        wheel_tag=wheel_tag,
    )
    wheel = builder.build(output_dir=output_dir, editable=editable)
    log.info(f"Built wheel: {wheel}")
    return wheel



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

import io
import logging
import os
import tarfile
from copy import copy
from gzip import GzipFile
from pathlib import Path
from posixpath import join as pjoin

from . import common

log = logging.getLogger(__name__)


def clean_tarinfo(ti, mtime=None):
    """Clean metadata from a TarInfo object to make it more reproducible.

    - Set uid & gid to 0
    - Set uname and gname to ""
    - Normalise permissions to 644 or 755
    - Set mtime if not None
    """
    ti = copy(ti)
    ti.uid = 0
    ti.gid = 0
    ti.uname = ""
    ti.gname = ""
    ti.mode = common.normalize_file_permissions(ti.mode)
    if mtime is not None:
        ti.mtime = mtime
    return ti


class SdistBuilder(common.FileSelector):
    """Builds an sdist"""

    def __init__(
        self,
        pyproject_file,
        metadata,
        entrypoints,
        includes=(),
        excludes=(),
        sdist_extra_includes=(),
        sdist_extra_excludes=(),
        metadata_files=(),
    ):
        pyproject_file = pyproject_file or "pyproject.toml"
        self.pyproject_file = Path(pyproject_file).absolute()
        self.base_dir = self.pyproject_file.parent
        self.metadata = metadata
        self.entrypoints = entrypoints

        self.include_patterns = common.FilePatterns(
            patterns=includes, base_dir=self.base_dir
        )
        self.exclude_patterns = common.FilePatterns(
            patterns=excludes, base_dir=self.base_dir
        )

        self.sdist_extra_include_patterns = common.FilePatterns(
            patterns=sdist_extra_includes, base_dir=self.base_dir
        )
        self.sdist_extra_exclude_patterns = common.FilePatterns(
            patterns=sdist_extra_excludes, base_dir=self.base_dir
        )

        # this is the only implict addition that is too easily overseen
        if not metadata_files:
            metadata_files = (
                "README*",
                "LICENSE*",
                "LICENCE*",
                "COPYING*",
            )
        self.metadata_file_patterns = common.FilePatterns(
            patterns=metadata_files, base_dir=self.base_dir
        )

    @classmethod
    def from_pyproject_file(
        cls,
        pyproject_file="pyproject.toml",
    ):
        """
        Create an sdist from an arbitrary pyproject.toml-like file path (default:
        pyproject.toml)
        """
        from .config import read_pyproject_file

        pyproject_file = pyproject_file or "pyproject.toml"
        project_info = read_pyproject_file(pyproject_file)

        return cls(
            pyproject_file=pyproject_file,
            metadata=common.Metadata(project_info.metadata),
            entrypoints=project_info.entrypoints,
            includes=project_info.includes,
            excludes=project_info.excludes,
            sdist_extra_includes=project_info.sdist_extra_includes,
            sdist_extra_excludes=project_info.sdist_extra_excludes,
            metadata_files=project_info.metadata_files,
        )

    @property
    def sdist_filename(self):
        return f"{self.dist_name}.tar.gz"

    @property
    def dist_name(self):
        return common.normalize_dist_name(self.metadata.name, self.metadata.version)

    def select_all_files(self):
        yield from self.select_files()
        yield from self.select_extra_sdist_files()
        yield from self.select_metafiles()
        # Always include the pyproject file at the root using standard name
        ppt_abs_path = self.pyproject_file.absolute()
        ppt_rel_path = Path("pyproject.toml")
        yield ppt_abs_path, ppt_rel_path

    def build(self, output_dir=None):
        """
        Build sdist and return sdist file path.
        """
        output_dir = (
            output_dir
            and Path(output_dir).absolute()
            or Path(self.pyproject_file.parent) / "dist"
        )
        output_dir.mkdir(parents=True, exist_ok=True)
        target = output_dir / self.sdist_filename

        source_date_epoch = os.environ.get("SOURCE_DATE_EPOCH", "")
        mtime = int(source_date_epoch) if source_date_epoch else None

        # For the gzip timestamp, default to 2016-1-1 00:00 (UTC)
        # This makes the sdist reproducible even without SOURCE_DATE_EPOCH,
        # if the source file mtimes don't change, i.e. from the same checkout.
        gz = GzipFile(str(target), mode="wb", mtime=(mtime or 1451606400))
        tar_file = tarfile.TarFile(
            str(target),
            mode="w",
            fileobj=gz,
            format=tarfile.PAX_FORMAT,
        )

        try:
            for abs_path, rel_path in self.select_all_files():
                self._add_file(
                    abs_path=abs_path,
                    tar_file=tar_file,
                    mtime=mtime,
                    rel_path=rel_path,
                )

            stream = io.StringIO()
            self.metadata.write_metadata_file(stream)
            pkg_info = stream.getvalue().encode()
            ti = tarfile.TarInfo(pjoin(self.dist_name, "PKG-INFO"))
            ti.size = len(pkg_info)
            tar_file.addfile(ti, io.BytesIO(pkg_info))

        finally:
            tar_file.close()
            gz.close()

        return target

    def _add_file(self, abs_path, tar_file, mtime, rel_path=None):
        if not rel_path:
            # If not provided, create the rel_path as relative to the pyproject
            # dir otherwise the includes/excludes will no longer work if trying
            # to build a wheel from an sdist
            rel_path = Path(abs_path).relative_to(self.pyproject_dir)

        log.debug("Adding {abs_path!r} to tar file as {rel_path!r}")
        abs_path, rel_path = str(abs_path), str(rel_path)
        ti = tar_file.gettarinfo(abs_path, arcname=pjoin(self.dist_name, rel_path))
        ti = clean_tarinfo(ti, mtime)

        if ti.isreg():
            with open(abs_path, "rb") as f:
                tar_file.addfile(ti, f)
        else:
            tar_file.addfile(ti)  # Symlinks & ?


def make_sdist(pyproject_file=None, output_dir=None):
    """
    Make an sdist, return the path.
    """
    builder = SdistBuilder.from_pyproject_file(pyproject_file=pyproject_file)
    sdist = builder.build(output_dir=output_dir)
    log.info(f"Built sdist: {sdist}")
    return sdist



#
# SPDX-License-Identifier: BSD-3-clause
# Copyright (c) nexB Inc. and contributors
# Copyright (c) 2015, Thomas Kluyver and contributors
# Based on https://github.com/pypa/flit/ and heavily modified

import errno
import logging
import os.path as osp
import re
from email.headerregistry import Address
from pathlib import Path

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from .versionno import normalize_version

log = logging.getLogger(__name__)


class ConfigError(ValueError):
    pass


pep621_allowed_fields = {
    "name",
    "version",
    "description",
    "readme",
    "requires-python",
    "license",
    "authors",
    "maintainers",
    "keywords",
    "classifiers",
    "urls",
    "scripts",
    "gui-scripts",
    "entry-points",
    "dependencies",
    "optional-dependencies",
}


def read_pyproject_file(pyproject_file=None):
    """
    Read and check the "pyproject.toml-like" pyproject_file project file.
    """
    pyproject_file = pyproject_file or "pyproject.toml"
    pyproject_file = Path(pyproject_file).absolute()
    pyproject = tomllib.loads(pyproject_file.read_text("utf-8"))
    return prep_pyproject_config(pyproject, pyproject_file)


def prep_pyproject_config(pyproject_data, path):
    """
    Return a ProjectInfo loaded from pyproject.toml and prepare common metadata
    """
    if not "project" in pyproject_data:
        raise ConfigError(f"[project] not found in {path}")

    project_info = read_pep621_metadata(pyproject_data["project"], path)

    flot = pyproject_data.get("tool", {}).get("flot", {})
    if not flot:
        raise ConfigError(f"[tool.flot] not found in {path}")

    flot_config = _get_flot_config(flot)
    for name, value in flot_config.items():
        if value:
            setattr(project_info, name, value)

    return project_info


def _get_flot_config(flot_config):
    """Return a mapping of config data found in a pyproject.toml [tool.flot] section."""
    known_flot_keys = {
        "includes",
        "excludes",
        "metadata_files",
        "wheel_path_prefixes_to_strip",
        "editable_paths",
        "sdist_extra_includes",
        "sdist_extra_excludes",
    }
    unknown_keys = set(flot_config) - known_flot_keys
    if unknown_keys:
        raise ConfigError(f"Unknown keys in [tool.flot]: " + ", ".join(unknown_keys))

    includes = flot_config.get("includes", [])
    if not includes:
        raise ConfigError("includes should contain at least one file or directory")
    includes = _check_glob_patterns(includes, "includes")

    base_excludes = ["**/.git/*", "**/.hg/*"]

    excludes = base_excludes + flot_config.get("excludes", [])
    excludes = _check_glob_patterns(excludes, "excludes")

    metadata_files = flot_config.get("metadata_files", [])
    metadata_files = _check_glob_patterns(metadata_files, "metadata_files")

    wheel_path_prefixes_to_strip = flot_config.get("wheel_path_prefixes_to_strip", [])
    if wheel_path_prefixes_to_strip:
        _check_list_of_str(flot_config, "wheel_path_prefixes_to_strip")

    editable_paths = flot_config.get("editable_paths", [])
    if editable_paths:
        _check_list_of_str(flot_config, "editable_paths")

    sdist_extra_includes = flot_config.get("sdist_extra_includes", [])
    sdist_extra_includes = _check_glob_patterns(
        sdist_extra_includes, "sdist_extra_includes"
    )

    sdist_extra_excludes = base_excludes + flot_config.get("sdist_extra_excludes", [])
    sdist_extra_excludes = _check_glob_patterns(
        sdist_extra_excludes, "sdist_extra_excludes"
    )

    return dict(
        includes=includes,
        excludes=excludes,
        metadata_files=metadata_files,
        wheel_path_prefixes_to_strip=wheel_path_prefixes_to_strip,
        editable_paths=editable_paths,
        sdist_extra_includes=sdist_extra_includes,
        sdist_extra_excludes=sdist_extra_excludes,
    )


def _check_glob_patterns(pats, clude):
    """Check and normalise a list of glob patterns (like for includes or excludes)"""
    if not isinstance(pats, list):
        raise ConfigError(f"{clude} patterns must be a list")

    # Windows filenames can't contain these (nor * or ?, but they are part of
    # glob patterns) - https://stackoverflow.com/a/31976060/434217
    bad_chars = re.compile(r'[\000-\037<>:"\\]')

    normed = []

    for p in pats:
        if bad_chars.search(p):
            raise ConfigError(
                '{} pattern {!r} contains bad characters (<>:"\\ or control characters)'.format(
                    clude, p
                )
            )

        normp = osp.normpath(p)

        if osp.isabs(normp):
            raise ConfigError("{} pattern {!r} is an absolute path".format(clude, p))
        if ".." in normp:
            raise ConfigError(
                "{} pattern {!r} contains relative .. and may point out of the base directory".format(
                    clude, p
                )
            )
        normed.append(normp)

    return normed


class ProjectInfo:
    def __init__(self):
        self.metadata = {}
        self.reqs_by_extra = {}
        self.entrypoints = {}
        self.referenced_files = []
        self.includes = []
        self.excludes = []
        self.metadata_files = []
        self.wheel_path_prefixes_to_strip = []
        self.editable_paths = []
        self.sdist_extra_includes = []
        self.sdist_extra_excludes = []

    def add_scripts(self, scripts_dict):
        if scripts_dict:
            self.entrypoints["console_scripts"] = scripts_dict

    def to_dict(self):
        return dict(
            metadata=self.metadata,
            reqs_by_extra=self.reqs_by_extra,
            entrypoints=self.entrypoints,
            referenced_files=self.referenced_files,
            includes=self.includes,
            excludes=self.excludes,
            metadata_files=self.metadata_files,
            wheel_path_prefixes_to_strip=self.wheel_path_prefixes_to_strip,
            editable_paths=self.editable_paths,
            sdist_extra_includes=self.sdist_extra_includes,
            sdist_extra_excludes=self.sdist_extra_excludes,
        )


readme_ext_to_content_type = {
    ".rst": "text/x-rst",
    ".md": "text/markdown",
    ".txt": "text/plain",
}


def description_from_file(rel_path: str, proj_dir: Path, guess_mimetype=True):
    if osp.isabs(rel_path):
        raise ConfigError("Readme path must be relative")

    desc_path = proj_dir / rel_path
    try:
        with desc_path.open("r", encoding="utf-8") as f:
            raw_desc = f.read()
    except IOError as e:
        if e.errno == errno.ENOENT:
            raise ConfigError("Description file {} does not exist".format(desc_path))
        raise

    if guess_mimetype:
        ext = desc_path.suffix.lower()
        try:
            mimetype = readme_ext_to_content_type[ext]
        except KeyError:
            log.warning("Unknown extension %r for description file.", ext)
            log.warning(
                "  Recognised extensions: %s", " ".join(readme_ext_to_content_type)
            )
            mimetype = None
    else:
        mimetype = None

    return raw_desc, mimetype


def _expand_requires_extra(re):
    for extra, reqs in sorted(re.items()):
        for req in reqs:
            if ";" in req:
                name, envmark = req.split(";", 1)
                yield '{} ; extra == "{}" and ({})'.format(name, extra, envmark)
            else:
                yield '{} ; extra == "{}"'.format(req, extra)


def _check_type(d, field_name, cls):
    if not isinstance(d[field_name], cls):
        raise ConfigError(
            "{} field should be {}, not {}".format(field_name, cls, type(d[field_name]))
        )


def _check_list_of_str(d, field_name):
    if not isinstance(d[field_name], list) or not all(
        isinstance(e, str) for e in d[field_name]
    ):
        raise ConfigError("{} field should be a list of strings".format(field_name))


def read_pep621_metadata(proj, path) -> ProjectInfo:
    lc = ProjectInfo()
    metadata = lc.metadata

    if "name" not in proj:
        raise ConfigError("name must be specified under [project]")
    _check_type(proj, "name", str)
    metadata["name"] = proj["name"]

    unexpected_keys = proj.keys() - pep621_allowed_fields
    if unexpected_keys:
        log.warning(f"Unexpected keys under [project]: {', '.join(unexpected_keys)}")

    if "version" not in proj:
        raise ConfigError("version must be specified under [project]")
    _check_type(proj, "version", str)
    metadata["version"] = normalize_version(proj["version"])

    if "description" not in proj:
        raise ConfigError("description must be specified under [project]")
    _check_type(proj, "description", str)
    metadata["summary"] = proj["description"]

    if "readme" in proj:
        readme = proj["readme"]
        if isinstance(readme, str):
            lc.referenced_files.append(readme)
            desc_content, mimetype = description_from_file(readme, path.parent)

        elif isinstance(readme, dict):
            unrec_keys = set(readme.keys()) - {"text", "file", "content-type"}
            if unrec_keys:
                raise ConfigError(
                    "Unrecognised keys in [project.readme]: {}".format(unrec_keys)
                )
            if "content-type" in readme:
                mimetype = readme["content-type"]
                mtype_base = mimetype.split(";")[0].strip()  # e.g. text/x-rst
                if mtype_base not in readme_ext_to_content_type.values():
                    raise ConfigError(
                        "Unrecognised readme content-type: {!r}".format(mtype_base)
                    )
                # TODO: validate content-type parameters (charset, md variant)?
            else:
                raise ConfigError(
                    "content-type field required in [project.readme] table"
                )
            if "file" in readme:
                if "text" in readme:
                    raise ConfigError(
                        "[project.readme] should specify file or text, not both"
                    )
                lc.referenced_files.append(readme["file"])
                desc_content, _ = description_from_file(
                    readme["file"], path.parent, guess_mimetype=False
                )
            elif "text" in readme:
                desc_content = readme["text"]
            else:
                raise ConfigError(
                    "file or text field required in [project.readme] table"
                )
        else:
            raise ConfigError("project.readme should be a string or a table")

        metadata["description"] = desc_content
        metadata["description_content_type"] = mimetype

    if "requires-python" in proj:
        metadata["requires_python"] = proj["requires-python"]

    if "license" in proj:
        _check_type(proj, "license", dict)
        license_tbl = proj["license"]
        unrec_keys = set(license_tbl.keys()) - {"text", "file"}
        if unrec_keys:
            raise ConfigError(
                "Unrecognised keys in [project.license]: {}".format(unrec_keys)
            )

        # TODO: Do something with license info.
        # The 'License' field in packaging metadata is a brief description of
        # a license, not the full text or a file path. PEP 639 will improve on
        # how licenses are recorded.
        if "file" in license_tbl:
            if "text" in license_tbl:
                raise ConfigError(
                    "[project.license] should specify file or text, not both"
                )
            lc.referenced_files.append(license_tbl["file"])
        elif "text" in license_tbl:
            pass
        else:
            raise ConfigError("file or text field required in [project.license] table")

    if "authors" in proj:
        _check_type(proj, "authors", list)
        metadata.update(pep621_people(proj["authors"]))

    if "maintainers" in proj:
        _check_type(proj, "maintainers", list)
        metadata.update(pep621_people(proj["maintainers"], group_name="maintainer"))

    if "keywords" in proj:
        _check_list_of_str(proj, "keywords")
        metadata["keywords"] = ",".join(proj["keywords"])

    if "classifiers" in proj:
        _check_list_of_str(proj, "classifiers")
        metadata["classifiers"] = proj["classifiers"]

    if "urls" in proj:
        _check_type(proj, "urls", dict)
        project_urls = metadata["project_urls"] = []
        for label, url in sorted(proj["urls"].items()):
            project_urls.append("{}, {}".format(label, url))

    if "entry-points" in proj:
        _check_type(proj, "entry-points", dict)
        for grp in proj["entry-points"].values():
            if not isinstance(grp, dict):
                raise ConfigError(
                    "projects.entry-points should only contain sub-tables"
                )
            if not all(isinstance(k, str) for k in grp.values()):
                raise ConfigError(
                    "[projects.entry-points.*] tables should have string values"
                )
        if set(proj["entry-points"].keys()) & {"console_scripts", "gui_scripts"}:
            raise ConfigError(
                "Scripts should be specified in [project.scripts] or "
                "[project.gui-scripts], not under [project.entry-points]"
            )
        lc.entrypoints = proj["entry-points"]

    if "scripts" in proj:
        _check_type(proj, "scripts", dict)
        if not all(isinstance(k, str) for k in proj["scripts"].values()):
            raise ConfigError("[project.scripts] table should have string values")
        lc.entrypoints["console_scripts"] = proj["scripts"]

    if "gui-scripts" in proj:
        _check_type(proj, "gui-scripts", dict)
        if not all(isinstance(k, str) for k in proj["gui-scripts"].values()):
            raise ConfigError("[project.gui-scripts] table should have string values")
        lc.entrypoints["gui_scripts"] = proj["gui-scripts"]

    if "dependencies" in proj:
        _check_list_of_str(proj, "dependencies")
        reqs_noextra = proj["dependencies"]
    else:
        reqs_noextra = []

    if "optional-dependencies" in proj:
        _check_type(proj, "optional-dependencies", dict)
        optdeps = proj["optional-dependencies"]
        if not all(isinstance(e, list) for e in optdeps.values()):
            raise ConfigError("Expected a dict of lists in optional-dependencies field")
        for e, reqs in optdeps.items():
            if not all(isinstance(a, str) for a in reqs):
                raise ConfigError(
                    f"Expected a string list for optional-dependencies ({e})"
                )

        lc.reqs_by_extra = optdeps.copy()
        metadata["provides_extra"] = sorted(lc.reqs_by_extra.keys())

    metadata["requires_dist"] = reqs_noextra + list(
        _expand_requires_extra(lc.reqs_by_extra)
    )

    if "dynamic" in proj:
        raise ConfigError("flot does not support dynamic fields")

    return lc


def pep621_people(people, group_name="author") -> dict:
    """Convert authors/maintainers from PEP 621 to core metadata fields"""
    names, emails = [], []
    for person in people:
        if not isinstance(person, dict):
            raise ConfigError("{} info must be list of dicts".format(group_name))
        unrec_keys = set(person.keys()) - {"name", "email"}
        if unrec_keys:
            raise ConfigError(
                "Unrecognised keys in {} info: {}".format(group_name, unrec_keys)
            )
        if "email" in person:
            email = person["email"]
            if "name" in person:
                email = str(Address(person["name"], addr_spec=email))
            emails.append(email)
        elif "name" in person:
            names.append(person["name"])

    res = {}
    if names:
        res[group_name] = ", ".join(names)
    if emails:
        res[group_name + "_email"] = ", ".join(emails)
    return res
